/**
 * @file **babel-plugin-deno (index.js)**
 * @description A simple babel plugin that rewrites
 * @author Brandon Kalinowski
 * @copyright 2020 Brandon Kalinowski
 * @copyright 2019 justavac (Helper Functions)
 * @license MIT
 */
//@ts-nocheck

const fs = require('fs');
const path = require('path');

module.export = function deno() {
  return {
    name: 'deno',
    visitor: {
      ImportDeclaration: {
        enter(path) {
          const source = path.node.source.value;
          path.node.source.value = convertRemoteToLocalCache(source);
        },
      },
    },
  };
};

/// Helpers ///

/**
 * strip Query
 * @param {strubg} moduleName
 */
function getModuleWithQueryString(moduleName) {
  let name = moduleName;
  for (
    const index = name.indexOf('?');
    index !== -1;
    name = name.substring(index + 1)
  ) {
    if (name.substring(0, index).endsWith('.ts')) {
      const cutLength = moduleName.length - name.length;
      return moduleName.substring(0, index + cutLength);
    }
  }
}

/**
 * strip extension
 * @param {string} moduleName
 * @returns {string}
 */
function stripExtNameDotTs(moduleName) {
  const moduleWithQuery = getModuleWithQueryString(moduleName);
  if (moduleWithQuery) {
    return moduleWithQuery;
  }

  if (!moduleName.endsWith('.ts')) {
    return moduleName;
  }

  const name = moduleName.slice(0, -3);
  logger.info(`strip "${moduleName}" to "${name}".`);

  return name;
}

/** @param {string} moduleName */
function convertRemoteToLocalCache(moduleName) {
  if (!moduleName.startsWith('http://') && !moduleName.startsWith('https://')) {
    return moduleName;
  }

  const denoDir = getDenoDir();
  // "https://deno.land/x/std/log/mod" to "$DENO_DIR/deps/https/deno.land/x/std/log/mod" (no ".ts" because stripped)
  const name = path.resolve(denoDir, 'deps', moduleName.replace('://', '/'));
  const redirectedName = fallbackHeader(name);
  logger.info(`convert "${moduleName}" to "${redirectedName}".`);

  return redirectedName;
}

/**
 * @template {{ mime_type: string; redirect_to: string; }} IDenoModuleHeaders
 */

/**
 * If moduleName is not found, recursively search for headers and "redirect_to" property.
 * @param {string}
 * @returns {string}
 */
function fallbackHeader(modulePath) {
  const validPath = modulePath.endsWith('.ts')
    ? modulePath
    : `${modulePath}.ts`;
  if (fs.existsSync(validPath)) {
    return modulePath;
  }

  const headersPath = `${validPath}.headers.json`;
  if (fs.existsSync(headersPath)) {
    /** @type IDenoModuleHeaders */
    const headers = JSON.parse(
      fs.readFileSync(headersPath, { encoding: 'utf-8' })
    );
    logger.info(`redirect "${modulePath}" to "${headers.redirect_to}".`);
    // TODO: avoid Circular
    return convertRemoteToLocalCache(stripExtNameDotTs(headers.redirect_to));
  }
  return modulePath;
}
